#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

# coding: utf-8

"""
     <i>Fecha de creación: 5 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""

import os
from collections import OrderedDict
from datetime import datetime, timedelta

from oauthlib.oauth2 import BackendApplicationClient, LegacyApplicationClient
from requests_oauthlib import OAuth2Session
from requests.auth import HTTPBasicAuth

from wire4_client import ApiClient
from wire4_client.configuration import Configuration
from wire4_client.rest import ApiException

from wire4_auth_p27.core.cached_token import CachedToken
from wire4_auth_p27.core.environment_enum import EnvironmentEnum


class OAuthWire4:

    MAX_APP_USER_SIZE_CACHED = 100

    def __init__(self, client_id, client_secret, environment=EnvironmentEnum.SANDBOX):
        os.environ["OAUTHLIB_RELAX_TOKEN_SCOPE"] = "True"
        self._client_id = client_id
        self._client_secret = client_secret
        self._environment = environment

        if self._environment is None:
            self._environment = EnvironmentEnum.SANDBOX

        # noinspection PyTypeChecker
        self._token_cached_app = CachedToken(None, None, None)
        self._tokens_cached_app_user = OrderedDict()

    def obtain_access_token_app(self, scope="general"):

        if self._token_cached_app.token is not None and scope in self._token_cached_app.token.get("scope") and \
                self.__is_expired(self._token_cached_app.token.get("expires_at")) and \
                self._token_cached_app.token.get("access_token") is not None:

            return self.__format_to_header(self._token_cached_app.token.get("access_token"))

        try:
            auth = HTTPBasicAuth(self._client_id, self._client_secret)
            scopes = [scope]
            client = BackendApplicationClient(client_id=self._client_id, scope=scopes)

            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(token_url=self._environment.token_url, auth=auth)
            self._token_cached_app.token = token

            return self.__format_to_header(token.get("access_token"))
        except Exception as ex:
            raise ApiException(reason="error to obtain token app: {0}".format(ex))

    def obtain_access_token_app_user(self, user_key, secret_key, scope="spei_admin"):

        key_search = user_key + scope
        token_cached = self.tokens_cached_app_user.get(key_search)
        if token_cached is not None and scope in token_cached.token.get("scope") and \
                self.__is_expired(token_cached.token.get("expires_at")) and \
                token_cached.token.get("access_token") is not None:

            return self.__format_to_header(token_cached.token.get("access_token"))

        try:
            auth = HTTPBasicAuth(self._client_id, self._client_secret)
            scopes = [scope]
            client = LegacyApplicationClient(client_id=self._client_id)

            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(token_url=self._environment.token_url, username=user_key,
                                      password=secret_key, auth=auth, scope=scopes)

            if len(self._tokens_cached_app_user) + 1 > self.MAX_APP_USER_SIZE_CACHED:
                for key in self._tokens_cached_app_user:
                    self._tokens_cached_app_user.pop(key)
                    break

            self._tokens_cached_app_user[key_search] = CachedToken(user_key, secret_key, token)

            return self.__format_to_header(token.get("access_token"))
        except Exception as ex:
            raise ApiException(reason="error to obtain token app user: {0}".format(ex))

    def regenerate_access_token_app(self, scope="general"):

        try:
            auth = HTTPBasicAuth(self._client_id, self._client_secret)
            scopes = [scope]
            client = BackendApplicationClient(client_id=self._client_id, scope=scopes)

            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(token_url=self._environment.token_url, auth=auth)
            self._token_cached_app.token = token

            return self.__format_to_header(token.get("access_token"))
        except Exception as ex:
            raise ApiException(reason="error to obtain token app: {0}".format(ex))

    def regenerate_access_token_app_user(self, user_key, secret_key, scope="spei_admin"):

        key_search = user_key + scope
        try:
            auth = HTTPBasicAuth(self._client_id, self._client_secret)
            scopes = [scope]
            client = LegacyApplicationClient(client_id=self._client_id)

            oauth = OAuth2Session(client=client)
            token = oauth.fetch_token(token_url=self._environment.token_url, username=user_key,
                                      password=secret_key, auth=auth, scope=scopes)

            if len(self._tokens_cached_app_user) + 1 > self.MAX_APP_USER_SIZE_CACHED:
                for key in self._tokens_cached_app_user:
                    self._tokens_cached_app_user.pop(key)
                    break

            self._tokens_cached_app_user[key_search] = CachedToken(user_key, secret_key, token)

            return self.__format_to_header(token.get("access_token"))
        except Exception as ex:
            raise ApiException(reason="error to obtain token app user: {0}".format(ex))

    def __is_expired(self, expires_at):

        return expires_at is not None and datetime.fromtimestamp(expires_at) > (datetime.now() - timedelta(minutes=5))

    def __format_to_header(self, token):

        return "Bearer " + token

    def get_default_api_client(self):
        # Configure OAuth2 access token for authorization
        configuration = Configuration()
        configuration.host = self._environment.service_url

        return ApiClient(configuration)



#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

# coding: utf-8

"""
     <i>Fecha de creación: 5 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""

from enum import Enum


class EnvironmentEnum(Enum):
    __order__ = 'SANDBOX DEVELOPMENT PRODUCTION'

    SANDBOX = 'https://sandbox-api.wire4.mx/token', 'https://sandbox-api.wire4.mx/wire4/1.0.0'

    DEVELOPMENT = 'https://development-api.wire4.mx/token', 'https://development-api.wire4.mx/wire4/1.0.0'

    PRODUCTION = 'https://api.wire4.mx/token', 'https://api.wire4.mx/wire4/1.0.0'

    def __new__(cls, token_url, service_url):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj.token_url = token_url
        obj.service_url = service_url
        obj._value_ = value
        return obj



#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

# coding: utf-8

"""
     <i>Fecha de creación: 9 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""
from __future__ import absolute_import

# import oauth2 components
from wire4_auth_p27.auth.oauth_wire4 import OAuthWire4

# import core components
from wire4_auth_p27.core.environment_enum import EnvironmentEnum
from wire4_auth_p27.core.cached_token import CachedToken

# import webHook verification signature components
from wire4_auth_p27.webhook_verification_signature.utils_compute import UtilsCompute



#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

# coding: utf-8

"""
     <i>Fecha de creación: 5 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""
from oauthlib.oauth2 import OAuth2Token


class CachedToken:

    def __init__(self, user_key, user_secret, token):

        self._user_key = user_key
        self._user_secret = user_secret
        self._token = token

    @property
    def user_key(self):

        return self._user_key

    @property
    def user_secret(self):

        return self._user_secret

    @property
    def token(self):

        return self._token

    @token.setter
    def token(self, token):
        self._token = token



# coding: utf-8
#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.



"""
     <i>Fecha de creación: 9 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""

from setuptools import setup, find_packages  # noqa: H301

NAME = "wire4-auth"
VERSION = "1.1.0"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["requests_oauthlib >= 1.3"]

setup(
    name=NAME,
    version=VERSION,
    description="Wire4Authenticator",
    author="Wire4",
    author_email="speiok-team@tcpip.tech",
    url="http://wire4.mx",
    keywords=["Swagger", "Wire4Authenticator"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
    Herramienta para autenticación de la API de Wire4  # noqa: E501
    """
)



# coding: utf-8
#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

"""
     <i>Fecha de creación: 31 de agosto, 2020</i>

     author: Juan Mandujano
     version: 1.0
"""

import unittest

from wire4_auth.auth.oauth_wire4 import OAuthWire4
from wire4_auth.core.environment_enum import EnvironmentEnum
from wire4_client.rest import ApiException


class TokensTest(unittest.TestCase):
    CLIENT_ID: str = "FxUWmqYGZuv8O1qjxstvIyJothMa"
    CLIENT_SECRET: str = "kjwbkrPVgXsnaUGzthj55dsFhx4a"
    USER_KEY: str = "071e2b59b354186b3a0158de493536@sandbox.wire4.mx"
    SECRET_KEY: str = "0d1e33e94604f01b4e00d2fcb6b48f"
    AMBIENT: EnvironmentEnum = EnvironmentEnum.SANDBOX

    def test_givenBadCredentials_should_raisesException(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, 'self.CLIENT_SECRET', self.AMBIENT)
        with self.assertRaises(ApiException):
            oauth_wire.obtain_access_token_app("general")

    def test_givenValidCredentials_should_returnToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        token = oauth_wire.obtain_access_token_app("general")
        self.assertIsNotNone(token, "Token shouldn't be None")

    def test_sameCredentials_should_returnSameToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        first_token = oauth_wire.obtain_access_token_app("general")
        second_token = oauth_wire.obtain_access_token_app("general")
        self.assertEqual(first_token, second_token, "Should be same token")

    def test_givenSomeScopes_should_returnSingleTokenForAllScopes(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        token = oauth_wire.obtain_access_token_app("general codi_report")
        self.assertIsNotNone(token, "Token shouldn't be None")

    def test_givenManyTokenWithDifferentScopesForTheSameUser_should_returnAsManyTokenAsScopesRequested(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        general_token = oauth_wire.obtain_access_token_app("general")
        report_token = oauth_wire.obtain_access_token_app("codi_report")
        both_tokens = oauth_wire.obtain_access_token_app("general codi_report")
        self.assertNotEqual(general_token, report_token, "Shouldn't be equals")
        self.assertNotEqual(general_token, both_tokens, "Shouldn't be equals")
        self.assertNotEqual(report_token, both_tokens, "Shouldn't be equals")

    def test_givenManyTokenWithDifferentScopesForTheSameUser_should_keepsInCache(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        general_token_first = oauth_wire.obtain_access_token_app("general")
        report_token_first = oauth_wire.obtain_access_token_app("codi_report")
        both_tokens_first = oauth_wire.obtain_access_token_app("general codi_report")
        general_token_second = oauth_wire.obtain_access_token_app("general")
        report_token_second = oauth_wire.obtain_access_token_app("codi_report")
        both_tokens_second = oauth_wire.obtain_access_token_app("general codi_report")
        self.assertEquals(general_token_first, general_token_second, 'should be equals')
        self.assertEquals(report_token_first, report_token_second, 'should be equals')
        self.assertEquals(both_tokens_first, both_tokens_second, 'should be equals')

    def test_whenRegenerateToken_should_replaceOldToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        general_token_first = oauth_wire.obtain_access_token_app("general")
        report_token_first = oauth_wire.obtain_access_token_app("codi_report")
        both_tokens_first = oauth_wire.obtain_access_token_app("general codi_report")
        general_token_second = oauth_wire.obtain_access_token_app("general")
        report_token_second = oauth_wire.obtain_access_token_app("codi_report")
        both_tokens_second = oauth_wire.obtain_access_token_app("general codi_report")
        self.assertEquals(general_token_first, general_token_second, 'should be equals')
        self.assertEquals(report_token_first, report_token_second, 'should be equals')
        self.assertEquals(both_tokens_first, both_tokens_second, 'should be equals')
        general_token_regenerated = oauth_wire.regenerate_access_token_app("general")
        report_token_regenerated = oauth_wire.regenerate_access_token_app("codi_report")
        both_tokens_regenerated = oauth_wire.regenerate_access_token_app("general codi_report")
        self.assertNotEqual(general_token_first, general_token_regenerated, 'should be equals')
        self.assertNotEqual(report_token_first, report_token_regenerated, 'should be equals')
        self.assertNotEqual(both_tokens_first, both_tokens_regenerated, 'should be equals')

    def test_givenBadUserCredentials_should_raisesException(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        with self.assertRaises(ApiException):
            oauth_wire.obtain_access_token_app_user(self.USER_KEY, 'self.SECRET_KEY', 'spei_admin')

    def test_givenValidUserCredentials_should_returnToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        token = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        self.assertIsNotNone(token, "Token shouldn't be None")

    def test_sameUserCredentials_should_returnSameToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        first_token = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        second_token = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        self.assertEqual(first_token, second_token, "Should be same token")

    def test_givenSomeUserScopes_should_returnSingleTokenForAllScopes(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        token = oauth_wire.obtain_access_token_app("spei_admin codi_report")
        self.assertIsNotNone(token, "Token shouldn't be None")

    def test_givenManyTokenWithDifferentScopesForTheSameUserCredentials_should_returnAsManyTokenAsScopesRequested(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        spei_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        spid_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spid_admin')
        codi_general = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_general')
        codi_report = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_report')
        self.assertNotEqual(spei_admin, spid_admin, "shouldn't be equals")
        self.assertNotEqual(spei_admin, codi_general, "shouldn't be equals")
        self.assertNotEqual(spei_admin, codi_report, "shouldn't be equals")
        self.assertNotEqual(codi_report, codi_general, "shouldn't be equals")
        self.assertNotEqual(spid_admin, codi_report, "shouldn't be equals")
        self.assertNotEqual(spid_admin, codi_general, "shouldn't be equals")

    def test_givenManyTokenWithDifferentScopesForTheSameUserCredentials_should_keepsInCache(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        spei_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        spid_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spid_admin')
        codi_general = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_general')
        codi_report = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_report')
        double_scope = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY,
                                                               'codi_report codi_general')
        spei_admin_second = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        spid_admin_second = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spid_admin')
        codi_general_second = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_general')
        codi_report_second = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_report')
        double_scope_second = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY,
                                                                      'codi_report codi_general')
        self.assertEqual(spei_admin, spei_admin_second, 'should be equals')
        self.assertEqual(spid_admin, spid_admin_second, 'should be equals')
        self.assertEqual(codi_general, codi_general_second, 'should be equals')
        self.assertEqual(codi_report, codi_report_second, 'should be equals')
        self.assertEqual(double_scope, double_scope_second, 'should be equals')

    def test_whenRegenerateTokenForUser_should_replaceOldToken(self):
        oauth_wire = OAuthWire4(self.CLIENT_ID, self.CLIENT_SECRET, self.AMBIENT)
        spei_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        spid_admin = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spid_admin')
        codi_general = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_general')
        codi_report = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_report')
        double_scope = oauth_wire.obtain_access_token_app_user(self.USER_KEY, self.SECRET_KEY,
                                                               'codi_report codi_general')
        spei_admin_second = oauth_wire.regenerate_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spei_admin')
        spid_admin_second = oauth_wire.regenerate_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'spid_admin')
        codi_general_second = oauth_wire.regenerate_access_token_app_user(self.USER_KEY, self.SECRET_KEY,
                                                                          'codi_general')
        codi_report_second = oauth_wire.regenerate_access_token_app_user(self.USER_KEY, self.SECRET_KEY, 'codi_report')
        double_scope_second = oauth_wire.regenerate_access_token_app_user(self.USER_KEY, self.SECRET_KEY,
                                                                          'codi_report codi_general')
        self.assertNotEqual(spei_admin, spei_admin_second, 'should be equals')
        self.assertNotEqual(spid_admin, spid_admin_second, 'should be equals')
        self.assertNotEqual(codi_general, codi_general_second, 'should be equals')
        self.assertNotEqual(codi_report, codi_report_second, 'should be equals')
        self.assertNotEqual(double_scope, double_scope_second, 'should be equals')


if __name__ == '__main__':
    unittest.main()



#    COPYRIGHT © 2017. TCPIP.
#    PATENT PENDING. ALL RIGHTS RESERVED.
#    SPEI GATEWAY IS REGISTERED TRADEMARKS OF TCPIP.
#
#    This software is confidential and proprietary information of TCPIP.
#    You shall not disclose such Confidential Information and shall use it only
#    in accordance with the company policy.

# coding: utf-8

"""
     <i>Fecha de creación: 9 de diciembre, 2019</i>

     author: Saintiago García
     version: 1.0
"""
import hashlib
import hmac


class UtilsCompute:
    DEFAULT_ENCODING = 'utf8'

    @staticmethod
    def compare_webhook_msg_signatures(message, key, header_signature):
        byte_array = bytearray(key, UtilsCompute.DEFAULT_ENCODING)
        sign = hmac.new(byte_array, message.encode(), hashlib.sha512).hexdigest()

        return sign == header_signature
